﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

namespace MapFold
{
    internal static class Extensions
    {
        public static IEnumerable<TResult> Map<TArg, TResult>(
            this IEnumerable<TArg> list,
            Func<TArg, TResult> func)
        {
            foreach (TArg item in list)
            {
                yield return func(item);
            }
        }

        public static IEnumerable<TResult> Repeat<TArg, TResult>
            (this TArg val, Func<TArg, TResult> func, int times)
        {
            for (int i = 0; i < times; i++)
            {
                yield return func(val);
            }
        }

        public static T Fold<T>(
            this IEnumerable<T> list,
            Func<T, T, T> func)
        {
            return Fold(list, func, default(T));
        }

        public static TResult Fold<TArg, TResult>(
            this IEnumerable<TArg> list,
            Func<TResult, TArg, TResult> func)
        {
            return Fold(list, func, default(TResult));
        }

        public static TResult Fold<TArg, TResult>(
            this IEnumerable<TArg> list,
            Func<TResult, TArg, TResult> func, TResult result)
        {
            foreach (TArg item in list)
            {
                result = func(result, item);
            }
            return result;
        }

        public static void ForEach<TArg>(
            this IEnumerable<TArg> list,
            Action<TArg> action)
        {
            foreach (TArg item in list)
            {
                action(item);
            }
        }

        public static IEnumerable<T> Apply<T>(this IEnumerable<T> list)
        {
            var result = new List<T>();
            foreach (T item in list)
            {
                result.Add(item);
            }
            return result;
        }

        public static IEnumerable<T> Filter<T>(
            this IEnumerable<T> list,
            Predicate<T> func)
        {
            foreach (T item in list)
            {
                if (func(item))
                {
                    yield return item;
                }
            }
        }

        public static int Count<T>(this IEnumerable<T> list)
        {
            int i = 0;
            foreach (T item in list)
            {
                i++;
            }
            return i;
        }

        public static IEnumerable<T[]> Zip<T>(this IEnumerable<T> list1, IEnumerable<T> list2)
        {
            var enumerators = new[] {list1.GetEnumerator(), list2.GetEnumerator()};           
            while (true)
            {
                int current = 0;
                var result = new T[enumerators.Length];
                foreach (IEnumerator<T> enumerator in enumerators)
                {
                    if (!enumerator.MoveNext())
                        yield break;
                    result[current++] = enumerator.Current;
                }
                yield return result;
            }
        }

        public static IEnumerable<TResult> ParallelMap<TArg, TResult>(
            this IEnumerable<TArg> list,
            Func<TArg, TResult> func)
        {
            var result = new TResult[list.Count()];
            using (var resetEvent = new ManualResetEvent(false))
            {
                int count = result.Length;                
                int i = 0;
                foreach (TArg item in list)
                {
                    TArg localItem = item;
                    int index = i;
                    ThreadPool.QueueUserWorkItem(
                            n =>
                            {                                
                                TResult value = func(localItem);
                                lock (result)
                                {
                                    result[index] = value;
                                }

                                if (Interlocked.Decrement(ref count) == 0)
                                {
                                    resetEvent.Set();
                                }
                            }                        
                        );
                    i++;
                }
                resetEvent.WaitOne();
            }
            return result;
        }        

        public static void WriteToConsole<T>(this IEnumerable<T> data, string label)
        {
            string result = string.Empty;
            foreach (T item in data)
            {
                result += item + ",";
            }
            result = result.TrimEnd(',');
            Console.WriteLine(label + result);
        }
    }
}
